"""
    QApp Platform Project oqc_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qcaas_client.client import OQCClient

from qapp_common.enum.provider_tag import ProviderTag
from qapp_common.model.provider.provider import Provider
from qapp_common.config.logging_config import logger


class OqcCloudProvider(Provider):
    def __init__(self, oqc_cloud_url: str, access_token: str):
        super().__init__(ProviderTag.OQC_CLOUD)
        self.oqc_cloud_url = oqc_cloud_url
        self.access_token = access_token

    def get_backend(self, device_specification):
        logger.debug("[OqcCloudProvider] get_backend()")

        return self.collect_provider()

    def collect_provider(self):
        logger.debug("[OqcCloudProvider] collect_provider()")

        return OQCClient(url=self.oqc_cloud_url, authentication_token=self.access_token)



"""
    QApp Platform Project braket_device_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qapp_common.enum.provider_tag import ProviderTag
from qapp_common.enum.sdk import Sdk
from qapp_common.factory.device_factory import DeviceFactory
from qapp_common.model.provider.provider import Provider
from qapp_common.config.logging_config import logger

from qapp_braket.model.device.aws_braket_device import AwsBraketDevice
from qapp_braket.model.device.oqc_cloud_device import OqcCloudDevice
from qapp_braket.model.device.qapp_braket_device import QappBraketDevice


class BraketDeviceFactory(DeviceFactory):

    @staticmethod
    def create_device(provider: Provider, device_specification: str, authentication: dict, sdk: Sdk):
        logger.info("[BraketDeviceFactory] create_device()")

        provider_type = ProviderTag.resolve(provider.get_provider_type().value)

        if ProviderTag.QUAO_QUANTUM_SIMULATOR.__eq__(provider_type) and Sdk.BRAKET.__eq__(sdk):
            logger.debug("[BraketDeviceFactory] Create QappBraketDevice")

            return QappBraketDevice(provider, device_specification)

        if ProviderTag.AWS_BRAKET.__eq__(provider_type):
            logger.debug("[BraketDeviceFactory] Create QappBraketDevice")

            return AwsBraketDevice(
                provider,
                device_specification,
                authentication.get("bucketName"),
                authentication.get("prefix"),
            )

        if ProviderTag.OQC_CLOUD.__eq__(provider_type):
            return OqcCloudDevice(provider, device_specification)

        raise Exception("Unsupported device!")



"""
    QuaO Project qapp_braket_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import time

from qapp_common.config.logging_config import logger
from qapp_common.data.device.circuit_running_option import CircuitRunningOption
from qapp_common.enum.status.job_status import JobStatus
from qapp_common.model.device.custom_device import CustomDevice
from qapp_common.model.provider.provider import Provider


class QappBraketDevice(CustomDevice):

    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[QappBraketDevice] _create_job()")

        start_time = time.time()

        job = self.device.run(task_specification=circuit, shots=options.shots)

        self.execution_time = time.time() - start_time

        return job

    def _produce_histogram_data(self, job_result) -> dict:
        logger.debug("[QappBraketDevice] _produce_histogram_data()")

        return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        logger.debug("[QappBraketDevice] _get_provider_job_id()")

        return job.id

    def _get_job_status(self, job) -> str:
        logger.debug("[QappBraketDevice] _get_job_status()")

        job_state = job.state()
        logger.debug("[AwsBraketDevice] job status: {0} ".format(job_state))

        if JobStatus.COMPLETED.value.__eq__(job_state):
            job_state = JobStatus.DONE.value

        return job_state

    def _is_simulator(self) -> bool:
        logger.debug("[QappBraketDevice] _is_simulator()")

        return True

    def _calculate_execution_time(self, job_result):
        logger.debug("[QappBraketDevice] _is_simulator()")

        logger.debug("[QappBraketDevice] Execution time calculation was: {0} seconds"
                     .format(self.execution_time))

    def _get_job_result(self, job):
        logger.debug("[QappBraketDevice] _get_job_result()")

        return job.result()

    def _get_shots(self, job_result) -> int | None:
        """
        Retrieve the number of shots from the job result.

        This method checks if the job result contains task metadata and
        retrieves the number of shots if available. If task metadata
        does not exist or does not contain the shots attribute,
        the method returns None.

        Args:
            job_result: An object representing the result of a job, which
                        may contain task metadata.

        Returns:
            int | None: The number of shots if available; otherwise, None.
        """
        logger.debug("[QuappBraketDevice] _get_shots()")

        return job_result.task_metadata.shots if hasattr(job_result, 'task_metadata') and hasattr(
            job_result.task_metadata, 'shots') else None



"""
    QApp Platform Project braket_circuit_export_task.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.circuits import Circuit

from qapp_common.async_tasks.export_circuit_task import CircuitExportTask
from qapp_common.config.logging_config import logger

from qapp_braket.util.circuit_convert_utils import CircuitConvertUtils


class BraketCircuitExportTask(CircuitExportTask):

    def _transpile_circuit(self):
        logger.debug("[BraketCircuitExportTask] _transpile_circuit()")

        circuit = self.circuit_data_holder.circuit

        if isinstance(circuit, Circuit):
            return CircuitConvertUtils.braket_to_qiskit(circuit)

        raise Exception("Invalid circuit type!")



"""
    QApp Platform Project braket_handler_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from qapp_common.config.logging_config import logger
from qapp_common.factory.handler_factory import HandlerFactory
from qapp_common.handler.handler import Handler

from ..handler.invocation_handler import InvocationHandler


class BraketHandlerFactory(HandlerFactory):

    @staticmethod
    def create_handler(event, circuit_preparation_fn, post_processing_fn) -> Handler:
        logger.info("[BraketHandlerFactory] create_handler()")

        request_data = event.json()

        logger.debug("[BraketHandlerFactory] Create InvocationHandler")
        return InvocationHandler(
            request_data=request_data,
            circuit_preparation_fn=circuit_preparation_fn,
            post_processing_fn=post_processing_fn,
        )



"""
    QApp Platform Project braket_invocation.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.circuits import Circuit

from qapp_common.component.backend.invocation import Invocation
from qapp_common.data.async_task.circuit_export.backend_holder import BackendDataHolder
from qapp_common.data.async_task.circuit_export.circuit_holder import CircuitDataHolder
from qapp_common.data.request.invocation_request import InvocationRequest
from qapp_common.model.provider.provider import Provider
from qapp_common.config.logging_config import logger
from qapp_common.config.thread_config import circuit_exporting_pool

from ...factory.braket_provider_factory import BraketProviderFactory
from ...factory.braket_device_factory import BraketDeviceFactory
from ...async_tasks.braket_circuit_export_task import BraketCircuitExportTask


class BraketInvocation(Invocation):

    def __init__(self, request_data: InvocationRequest):
        super().__init__(request_data)

    def _export_circuit(self, circuit):
        logger.info("[BraketInvocation] _export_circuit()")

        circuit_export_task = BraketCircuitExportTask(
            circuit_data_holder=CircuitDataHolder(circuit, self.circuit_export_url),
            backend_data_holder=BackendDataHolder(
                self.backend_information, self.authentication.user_token
            ),
        )

        circuit_exporting_pool.submit(circuit_export_task.do)

    def _create_provider(self):
        logger.info("[BraketInvocation] _create_provider()")

        return BraketProviderFactory.create_provider(
            provider_type=self.backend_information.provider_tag,
            sdk=self.sdk,
            authentication=self.backend_information.authentication,
        )

    def _create_device(self, provider: Provider):
        logger.info("[BraketInvocation] _create_device()")

        return BraketDeviceFactory.create_device(
            provider=provider,
            device_specification=self.backend_information.device_name,
            authentication=self.backend_information.authentication,
            sdk=self.sdk,
        )

    def _get_qubit_amount(self, circuit):
        logger.info("[BraketInvocation] _get_qubit_amount()")

        if isinstance(circuit, Circuit):
            return circuit.qubit_count

        raise Exception("Invalid circuit type!")



"""
    QApp Platform Project braket_provider_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qapp_common.enum.provider_tag import ProviderTag
from qapp_common.enum.sdk import Sdk
from qapp_common.factory.provider_factory import ProviderFactory
from qapp_common.config.logging_config import logger

from ..model.provider.aws_braket_provider import AwsBraketProvider
from ..model.provider.oqc_cloud_provider import OqcCloudProvider
from ..model.provider.qapp_braket_provider import QappBraketProvider


class BraketProviderFactory(ProviderFactory):

    @staticmethod
    def create_provider(provider_type: ProviderTag, sdk: Sdk, authentication: dict):
        logger.info("[BraketProviderFactory] create_provider()")

        if ProviderTag.QUAO_QUANTUM_SIMULATOR.__eq__(provider_type) and Sdk.BRAKET.__eq__(sdk):
            logger.debug("[BraketProviderFactory] Create QappBraketProvider")

            return QappBraketProvider()

        if ProviderTag.AWS_BRAKET.__eq__(provider_type):
            logger.debug("[BraketProviderFactory] Create AwsBraketProvider")

            return AwsBraketProvider(
                authentication.get("accessKey"),
                authentication.get("secretKey"),
                authentication.get("regionName"),
            )

        if ProviderTag.OQC_CLOUD.__eq__(provider_type):
            return OqcCloudProvider(
                authentication.get("url"),
                authentication.get("accessToken")
            )

        raise Exception("Unsupported provider!")



"""
    QApp Platform Project circuit_convert_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.circuits import Circuit
from qiskit import QuantumCircuit
from qiskit.qasm2 import loads
from pytket.qasm import circuit_to_qasm_str
from pytket.extensions.braket import braket_to_tk

from qapp_common.config.logging_config import logger


class CircuitConvertUtils:

    @staticmethod
    def braket_to_qasm2_str(circuit: Circuit) -> str:
        logger.debug("[CircuitConvertUtils] braket_to_qasm2_str()")

        return circuit_to_qasm_str(braket_to_tk(circuit))

    @staticmethod
    def qasm2_str_to_qiskit(qasm_str: str) -> QuantumCircuit:
        logger.debug("[CircuitConvertUtils] qasm2_str_to_qiskit()")

        return loads(qasm_str)

    @staticmethod
    def braket_to_qiskit(circuit: Circuit) -> QuantumCircuit:
        logger.debug("[CircuitConvertUtils] braket_to_qiskit()")

        qasm_str = CircuitConvertUtils.braket_to_qasm2_str(circuit)
        return CircuitConvertUtils.qasm2_str_to_qiskit(qasm_str)



"""
    QApp Platform Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from time import time

from qapp_common.config.logging_config import logger
from qapp_common.data.device.circuit_running_option import CircuitRunningOption
from qapp_common.enum.status.job_status import JobStatus
from qapp_common.model.device.custom_device import CustomDevice
from qapp_common.model.provider.provider import Provider
from qcaas_client.client import QPUTask, CompilerConfig
from qiskit.qasm2 import dumps

from qapp_braket.util.circuit_convert_utils import CircuitConvertUtils


class OqcCloudDevice(CustomDevice):

    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)

        self.device_specification = device_specification

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[OqcCloudDevice] _create_job() with {0} shots".format(options.shots))

        start_time = time()

        qiskit_circuit = CircuitConvertUtils.braket_to_qiskit(circuit)
        qiskit_circuit.measure_all()
        qasm_str = dumps(qiskit_circuit)
        circuit_submit_options = CompilerConfig(repeats=options.shots)

        task = QPUTask(program=qasm_str, config=circuit_submit_options)
        job = self.device.execute_tasks(task, qpu_id=self.device_specification)

        self.execution_time = time() - start_time

        return job

    def _is_simulator(self) -> bool:
        logger.debug("[OqcCloudDevice] _is_simulator()")

        return True

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.debug("[OqcCloudDevice] _produce_histogram_data()")

        return next(iter(job_result.result.values()))

    def _get_provider_job_id(self, job) -> str:
        logger.debug("[OqcCloudDevice] _get_provider_job_id()")

        return job[0].id

    def _get_job_status(self, job) -> str:
        logger.debug("[OqcCloudDevice] _get_job_status()")

        oqc_status = self.device.get_task_status(task_id=job[0].id,
                                                 qpu_id=self.device_specification)
        logger.debug("[OqcCloudDevice] job status: {0} ".format(oqc_status))

        if "FAILED".__eq__(oqc_status):
            return JobStatus.ERROR.value
        elif "COMPLETED".__eq__(oqc_status):
            return JobStatus.DONE.value

        return oqc_status

    def _calculate_execution_time(self, job_result):
        logger.debug("[OqcCloudDevice] _calculate_execution_time()")

        logger.debug("[OqcCloudDevice] Execution time calculation was: {0} seconds"
                     .format(self.execution_time))

    def _get_job_result(self, job):
        logger.debug("[OqcCloudDevice] _get_job_result()")

        return job[0]

    def _get_shots(self, job_result) -> int | None:
        """
        Retrieve the total number of measurement shots from the job result.

        This method checks if the job result contains measurement data
        and calculates the total number of shots based on the measurement
        results. If the measurement data is not available, the method
        returns None.

        Args:
            job_result: An object representing the result of a job, which
                        may contain measurement results under the 'result'
                        attribute.

        Returns:
            int | None: The total number of shots if measurement data is
                         available; otherwise, None.
        """
        logger.debug("[OqcCloudDevice] _get_shots()")

        result_meas = getattr(job_result, 'result', {}).get('meas')
        return sum(result_meas.values()) if result_meas else None



"""
    QuaO Project qapp_braket_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from braket.devices import LocalSimulator

from qapp_common.config.logging_config import logger
from qapp_common.enum.provider_tag import ProviderTag
from qapp_common.model.provider.provider import Provider


class QappBraketProvider(Provider):

    def __init__(self, ):
        super().__init__(ProviderTag.QUAO_QUANTUM_SIMULATOR)

    def get_backend(self, device_specification):
        logger.debug('[QappBraketProvider] get_backend()')

        provider = self.collect_provider()

        device_names = provider.registered_backends()

        if device_names.__contains__(device_specification):
            return LocalSimulator(device_specification)

        raise Exception('Unsupported device')

    def collect_provider(self):
        logger.debug('[QappBraketProvider] collect_provider()')

        return LocalSimulator()



"""
    QuaO Project aws_braket_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from dateutil.parser import parse

from qapp_common.config.logging_config import logger
from qapp_common.data.device.circuit_running_option import CircuitRunningOption
from qapp_common.enum.status.job_status import JobStatus
from qapp_common.model.device.custom_device import CustomDevice
from qapp_common.model.provider.provider import Provider


class AwsBraketDevice(CustomDevice):

    def __init__(self, provider: Provider,
                 device_specification: str,
                 s3_bucket_name: str,
                 s3_prefix: str):
        super().__init__(provider, device_specification)
        self.s3_folder = (s3_bucket_name, s3_prefix)

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.debug("[AwsBraketDevice] _produce_histogram_data()")

        return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        logger.debug("[AwsBraketDevice] _get_provider_job_id()")

        return job.id

    def _get_job_status(self, job) -> str:
        logger.debug("[AwsBraketDevice] _get_job_status()")

        job_state = job.state()
        logger.debug("[AwsBraketDevice] job status: {0} ".format(job_state))

        if JobStatus.COMPLETED.value.__eq__(job_state):
            job_state = JobStatus.DONE.value
        return job_state

    def _calculate_execution_time(self, job_result):
        logger.debug("[AwsBraketDevice] _calculate_execution_time()")

        if "task_metadata" not in job_result:
            return

        task_metadata = job_result["task_metadata"]

        if task_metadata is None \
                or not bool(task_metadata) \
                or "createdAt" not in task_metadata \
                or "endedAt" not in task_metadata:
            return

        created_at = task_metadata["createdAt"]
        ended_at = task_metadata["endedAt"]

        if created_at is None or ended_at is None:
            return

        created_at = parse(created_at.replace("T", " ").replace("Z", ""))
        ended_at = parse(ended_at.replace("T", " ").replace("Z", ""))

        offset = ended_at - created_at

        self.execution_time = offset.total_seconds()

        logger.debug("[AwsBraketDevice] Execution time calculation was: {0} seconds"
                     .format(self.execution_time))

    def _get_job_result(self, job):
        logger.debug("[AwsBraketDevice] _get_job_result()")

        return job.result()

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[AwsBraketDevice] _create_job()")

        job = self.device.run(task_specification=circuit,
                              s3_destination_folder=self.s3_folder,
                              shots=options.shots)
        return job

    def _is_simulator(self) -> bool:
        logger.debug("[AwsBraketDevice] _is_simulator()")

        return True

    def _get_shots(self, job_result) -> int | None:
        """
        Retrieve the number of shots from the job result's task metadata.

        This method checks if the job result contains task metadata and
        retrieves the number of shots if available. If task metadata
        does not exist or does not contain the shots attribute,
        the method returns None.

        Args:
            job_result: An object representing the result of a job, which
                        may contain task metadata with the number of shots.

        Returns:
            int | None: The number of shots if available; otherwise, None.
        """
        logger.debug("[AwsBraketDevice] _get_shots()")

        return job_result.task_metadata.shots if hasattr(job_result, 'task_metadata') and hasattr(
            job_result.task_metadata, 'shots') else None
